"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from functools import reduce
from typing import Generic, TypeVar


__all__ = ["Monoid"]


m = TypeVar("m")

class Monoid(Generic[m], metaclass=ABCMeta):
    """ Abstract base class for monoids """
    @staticmethod
    @abstractmethod
    def mempty() -> Monoid[m]:
        """ mempty :: m """

    @abstractmethod
    def mappend(self, rhs:Monoid[m]) -> Monoid[m]:
        """ mappend :: m -> m -> m """

    @classmethod
    def mconcat(cls, *ms:Monoid[m]) -> Monoid[m]:
        """ mconcat :: [m] -> m """
        folder = lambda x, y: x.mappend(y)
        return reduce(folder, ms, cls.mempty())



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, Generic, TypeVar


__all__ = ["Monad"]


a = TypeVar("a")
b = TypeVar("b")


class Monad(Generic[a], metaclass=ABCMeta):
    """ Abstract base class for monads """
    @classmethod
    @abstractmethod
    def retn(cls, v:a) -> Monad[a]:
        """ return :: a -> m a """

    @abstractmethod
    def bind(self, rhs:Callable[[a], Monad[b]]) -> Monad[b]:
        """ (>>=) :: m a -> (a -> m b) -> m b """

    def __ge__(self, rhs:Callable[[a], Monad[b]]) -> Monad[b]:
        """ We use `x >= y` as Haskell's `x >>= y`; beware precedence rules """
        return self.bind(rhs)

    def then(self, rhs:Monad[b]) -> Monad[b]:
        """ (>>) :: m a -> m b -> m b """
        return self.bind(lambda _: rhs)

    def __rshift__(self, rhs:Monad[b]) -> Monad[b]:
        return self.then(rhs)

    @staticmethod
    def fail(msg:str) -> Monad[a]:
        """ fail :: String -> m a """
        raise Exception(msg)

    # TODO do notation analogue...maybe with context managers



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

import unittest
from typing import Generic, TypeVar

from finkl.abc import Eq, Monoid
from finkl.monad import List, Writer
from finkl.monoid import Sum


a = TypeVar("a")
m = TypeVar("m", bound=Monoid)

class _EqWriter(Generic[a, m], Writer[a, m], Eq):
    def __eq__(self, rhs:_EqWriter[a, m]) -> bool:
        return self.run_writer() == rhs.run_writer()

class _Logger(_EqWriter[int, List[str]]):
    writer = List

class _Counter(_EqWriter[str, Sum]):
    writer = Sum


class TestWriter_Logger(unittest.TestCase):
    def test_return(self):
        logger = _Logger.retn(123)
        self.assertEqual(logger, _Logger(123, List()))

    def test_bind(self):
        _inc = lambda x: _Logger(x + 1, List(f"Incremented {x}"))
        _dbl = lambda x: _Logger(x * 2, List(f"Doubled {x}"))

        result = _Logger(0).bind(_inc) \
                           .bind(_inc) \
                           .bind(_dbl)

        self.assertEqual(result, _Logger(4, List(
            "Incremented 0",
            "Incremented 1",
            "Doubled 2"
        )))

    def test_laws(self):
        # NOTE Obviously here we're just testing arbitrary values,
        # rather than over the entire type space.
        _inc = lambda x: _Logger(x + 1, List("foo"))
        _dbl = lambda x: _Logger(x * 2, List("bar"))

        # Left Identity
        self.assertEqual(_Logger.retn(123) >= _inc, _inc(123))

        # Right Identity
        self.assertEqual(_Logger(123) >= _Logger.retn, _Logger(123))

        ## Associativity
        self.assertEqual((_Logger(123) >= _inc) >= _dbl,
                         _Logger(123) >= (lambda x: _inc(x) >= _dbl))


class TestWriter_Counter(unittest.TestCase):
    def test_return(self):
        counter = _Counter.retn("Hello, World!")
        self.assertEqual(counter, _Counter("Hello, World!", Sum(0)))

    def test_bind(self):
        def _say(what):
            return lambda x: _Counter(f"{x}{what}", Sum(1))

        result = _Counter("Hello").bind(_say(", ")) \
                                  .bind(_say("World")) \
                                  .bind(_say("!"))

        self.assertEqual(result, _Counter("Hello, World!", Sum(3)))

    def test_laws(self):
        # NOTE Obviously here we're just testing arbitrary values,
        # rather than over the entire type space.
        _emph = lambda x: _Counter(f"{x}!", Sum(1))
        _qstn = lambda x: _Counter(f"{x}?", Sum(1))

        # Left Identity
        self.assertEqual(_Counter.retn("foo") >= _emph, _emph("foo"))

        # Right Identity
        self.assertEqual(_Counter("foo") >= _Counter.retn, _Counter("foo"))

        ## Associativity
        self.assertEqual((_Counter("foo") >= _emph) >= _qstn,
                         _Counter("foo") >= (lambda x: _emph(x) >= _qstn))


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from typing import Callable, TypeVar


__all__ = ["identity", "compose"]


a = TypeVar("a")
b = TypeVar("b")
c = TypeVar("c")


def identity(x:a) -> a:
    """ id :: a -> a """
    return x


def compose(f:Callable[[b], c], g:Callable[[a], b]) -> Callable[[a], c]:
    """ (.) :: (b -> c) -> (a -> b) -> (a -> c) """
    return lambda x: f(g(x))



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod


class Eq(metaclass=ABCMeta):
    """ Abstract base class for equality checking """
    @abstractmethod
    def __eq__(self, rhs:Eq) -> bool:
        """ Equality """

    def __neq__(self, rhs:Eq) -> bool:
        """ Inequality """
        return not self == rhs



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, Generic, TypeVar


__all__ = ["Functor", "Applicative"]


a = TypeVar("a")
b = TypeVar("b")


class Functor(Generic[a], metaclass=ABCMeta):
    """ Abstract base class for functors """
    @abstractmethod
    def fmap(self, fn:Callable[[a], b]) -> Functor[b]:
        """ fmap :: f a -> (a -> b) -> f b """


class Applicative(Generic[a, b], Functor[Callable[[a], b]], metaclass=ABCMeta):
    """ Abstract base class for applicative functors """
    @classmethod
    @abstractmethod
    def pure(cls, v:a) -> Functor[a]:
        """ pure :: a -> f a """

    @abstractmethod
    def applied_over(self, rhs:Functor[a]) -> Functor[b]:
        """ (<*>) :: f (a -> b) -> f a -> f b """

    def __matmul__(self, rhs:Functor[a]) -> Functor[b]:
        """ We use `x @ y` as Haskell's `x <*> y` """
        return self.applied_over(rhs)



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

import unittest

from finkl.monad import Maybe, Just, Nothing
from finkl.util import compose


_inc = lambda x: x + 1
_dbl = lambda x: x * 2


class TestEq(unittest.TestCase):
    def test_equality(self):
        self.assertEqual(Just(123), Just(123))
        self.assertEqual(Nothing, Nothing)
        self.assertNotEqual(Just(1), Just(2))
        self.assertNotEqual(Just(123), Nothing)
        self.assertNotEqual(Nothing, Just(123))


class TestFunctor(unittest.TestCase):
    def test_fmap(self):
        self.assertEqual(Just(123).fmap(_inc), Just(124))
        self.assertEqual(Nothing.fmap(_inc), Nothing)


class TestApplicative(unittest.TestCase):
    def test_pure(self):
        self.assertEqual(Maybe(123), Just(123))

    def test_applied_over(self):
        self.assertEqual(Just(_dbl) @ Just(123), Just(246))
        self.assertEqual(Just(_dbl) @ Nothing, Nothing)
        self.assertEqual(Nothing @ Just(123), Nothing)


class TestMonad(unittest.TestCase):
    def test_return(self):
        self.assertEqual(Maybe.retn(123), Just(123))

    def test_bind(self):
        self.assertEqual(Nothing >= compose(Just, _inc), Nothing)
        self.assertEqual(Nothing.bind(compose(Just, _inc)), Nothing)

        self.assertEqual(Just(123) >= compose(Just, _inc), Just(124))
        self.assertEqual(Just(123).bind(compose(Just, _inc)), Just(124))

        self.assertEqual((Just(123) >= compose(Just, _inc)) >= compose(Just, _dbl), Just(248))
        self.assertEqual(Just(123).bind(compose(Just, _inc)).bind(compose(Just, _dbl)), Just(248))

    def test_sequence(self):
        self.assertEqual(Just(123) >> Nothing, Nothing)
        self.assertEqual(Just(123).then(Nothing), Nothing)

    def test_complex(self):
        self.assertEqual((Just(123) >= compose(Just, _inc)) >> Just(1) >= compose(Just, _dbl), Just(2))

    def test_laws(self):
        # NOTE Obviously here we're just testing arbitrary values,
        # rather than over the entire type space.

        # Left Identity
        self.assertEqual(Maybe.retn(123) >= compose(Just, _inc), compose(Just, _inc)(123))

        # Right Identity
        self.assertEqual(Just(123) >= Maybe.retn, Just(123))

        # Associativity
        self.assertEqual((Just(123) >= compose(Just, _inc)) >= compose(Just, _dbl),
                         Just(123) >= (lambda x: Just(_inc(x)) >= compose(Just, _dbl)))

    def test_do_notation(self):
        self.assertEqual(Just("Hello") >= (lambda x: Just("World") >= (lambda y: Maybe(f"{x} {y}!"))), Just("Hello World!"))
        self.assertEqual(Just("Hello").bind(lambda x: Just("World").bind(lambda y: Maybe(f"{x} {y}!"))), Just("Hello World!"))


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

import unittest

from finkl.monad import List


class TestList(unittest.TestCase):
    def test_equality(self):
        self.assertNotEqual(List(1, 2, 3), List())
        self.assertEqual(List(1, 2, 3), List(1, 2, 3))

    def test_fmap(self):
        self.assertEqual(List(1, 2, 3).fmap(lambda x: x + 1), List(2, 3, 4))

    def test_return(self):
        self.assertEqual(List.retn(123), List(123))

    def test_bind(self):
        _inc = lambda x: List(x + 1)
        _dbl = lambda x: List(x * 2)

        result = List(1, 2, 3).bind(_inc).bind(_inc).bind(_dbl)
        self.assertEqual(result, List(6, 8, 10))

    def test_laws(self):
        # NOTE Obviously here we're just testing arbitrary values,
        # rather than over the entire type space.
        _inc = lambda x: List(x + 1)
        _dbl = lambda x: List(x * 2)

        # Left Identity
        self.assertEqual(List.retn(123) >= _inc, _inc(123))

        # Right Identity
        self.assertEqual(List(123) >= List.retn, List(123))

        ## Associativity
        self.assertEqual((List(123) >= _inc) >= _dbl,
                         List(123) >= (lambda x: _inc(x) >= _dbl))

    def test_mconcat(self):
        self.assertEqual(List.mconcat(), List())
        self.assertEqual(List.mconcat(*[List(x) for x in [1, 2, 3]]), List(1, 2, 3))


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from typing import Callable, Generic, List as _List, TypeVar

from finkl.abc import Eq, Functor, Monad, Monoid


__all__ = ["List"]


a = TypeVar("a")
b = TypeVar("b")

class List(Generic[a], Eq, Functor[_List[a]], Monad[_List[a]], Monoid[_List[a]]):
    """ List functor, monad and monoid """
    _v:_List[a]

    def __init__(self, *vs:a) -> None:
        self._v = list(vs)

    def __eq__(self, rhs:List[a]) -> bool:
        return self._v == rhs._v

    def __repr__(self) -> str:
        return repr(self._v)

    def fmap(self, fn:Callable[[a], b]) -> List[b]:
        return List(*[fn(v) for v in self._v])

    @classmethod
    def retn(cls, v:a) -> List[a]:
        return List(v)

    def bind(self, rhs:Callable[[a], List[b]]) -> List[b]:
        return List.mconcat(*self.fmap(rhs)._v)

    @staticmethod
    def fail(msg:str) -> List[a]:
        return List()

    @staticmethod
    def mempty():
        return List()

    def mappend(self, rhs:List[a]) -> List[a]:
        return List(*(self._v + rhs._v))



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from typing import Callable, ClassVar, Generic, Optional, Tuple, Type, TypeVar

from finkl.abc import Monad, Monoid


__all__ = ["Writer"]


a = TypeVar("a")
b = TypeVar("b")
m = TypeVar("m", bound=Monoid)

class Writer(Generic[a, m], Monad[a]):
    """ Writer monad """
    _v:a
    _w:m

    writer:ClassVar[Type[m]]

    def __init__(self, v:a, w:Optional[m] = None) -> None:
        self._v = v
        self._w = self.writer.mempty() if w is None else w

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._v)}, {repr(self._w)})"

    def run_writer(self) -> Tuple[a, m]:
        """ runWriter :: Writer w a -> (a, w) """
        return self._v, self._w

    @classmethod
    def retn(cls, v:a) -> Writer[a, m]:
        return cls(v)

    def bind(self, rhs:Callable[[a], Writer[b, m]]) -> Writer[b, m]:
        new = rhs(self._v)
        return self.__class__(new._v, self._w.mappend(new._w))



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta
from numbers import Number
from typing import Generic, TypeVar

from finkl.abc import Eq, Monoid


__all__ = ["Sum", "Product", "Any", "All"]


a = TypeVar("a")
m = TypeVar("m")


class _BaseMonoid(Generic[m], Eq, Monoid[m], metaclass=ABCMeta):
    """ We use inheritance to avoid too much boilerplate """
    _m:m

    def __init__(self, value:m) -> None:
        self._m = value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self._m)})"

    def __eq__(self, rhs:_BaseMonoid) -> bool:
        return self._m == rhs._m


class Sum(_BaseMonoid[Number]):
    """ Numeric sum monoid """
    @staticmethod
    def mempty():
        return Sum(0)

    def mappend(self, rhs:Sum) -> Sum:
        return Sum(self._m + rhs._m)


class Product(_BaseMonoid[Number]):
    """ Numeric product monoid """
    @staticmethod
    def mempty():
        return Product(1)

    def mappend(self, rhs:Product) -> Product:
        return Product(self._m * rhs._m)


class Any(_BaseMonoid[bool]):
    """ Any monoid """
    @staticmethod
    def mempty():
        return Any(False)

    def mappend(self, rhs:Any) -> Any:
        return Any(self._m or rhs._m)


class All(_BaseMonoid[bool]):
    """ All monoid """
    @staticmethod
    def mempty():
        return All(True)

    def mappend(self, rhs:All) -> All:
        return All(self._m and rhs._m)



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

import unittest

from finkl.monoid import Sum, Product, Any, All


class TestSum(unittest.TestCase):
    def test_mconcat(self):
        self.assertEqual(Sum.mconcat(), Sum(0))
        self.assertEqual(Sum.mconcat(*[Sum(x) for x in [1, 2, 3]]), Sum(6))


class TestProduct(unittest.TestCase):
    def test_mconcat(self):
        self.assertEqual(Product.mconcat(), Product(1))
        self.assertEqual(Product.mconcat(*[Product(x) for x in [1, 2, 3]]), Product(6))


class TestAny(unittest.TestCase):
    def test_mconcat(self):
        self.assertEqual(Any.mconcat(), Any(False))
        self.assertEqual(Any.mconcat(*[Any(x) for x in [False, False, True]]), Any(True))


class TestAll(unittest.TestCase):
    def test_mconcat(self):
        self.assertEqual(All.mconcat(), All(True))
        self.assertEqual(All.mconcat(*[All(x) for x in [True, True, True]]), All(True))


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from typing import Any, Generic, Type, TypeVar, Union

from finkl.abc import Eq, Applicative, Monad


__all__ = ["Maybe", "Just", "Nothing"]


a = TypeVar("a")
b = TypeVar("b")


class _Nothing:
    """ Nothing sentinel class """

class Maybe(Generic[a, b], Eq, Applicative[a, b], Monad[a]):
    """ Maybe """
    _v:Union[a, Type[_Nothing]]

    def __init__(self, v:Union[a, Type[_Nothing]]):
        self._v = v

    def __repr__(self):
        return "Nothing" if self._nothing else f"Just({repr(self._v)})"

    def __eq__(self, rhs:Maybe[a, Any]) -> bool:
        return self._v == rhs._v

    @property
    def _nothing(self):
        return self._v is _Nothing

    def fmap(self, fn:Callable[[a], b]) -> Maybe[b, Any]:
        return Maybe(_Nothing if self._nothing else fn(self._v))

    @classmethod
    def pure(cls, v:a) -> Maybe[a, Any]:
        return Maybe(v)

    def applied_over(self, rhs:Maybe[a, Any]) -> Maybe[b, Any]:
        return Maybe(_Nothing) if self._nothing else rhs.fmap(self._v)

    retn = pure

    def bind(self, rhs:Callable[[a], Maybe[b, Any]]) -> Maybe[b, Any]:
        return Maybe(_Nothing) if self._nothing else rhs(self._v)

    @staticmethod
    def fail(msg:str) -> Maybe[a, Any]:
        return Maybe(_Nothing)

Just = Maybe.pure
Nothing = Maybe(_Nothing)
